# HTTP CODES
HTTP_NO_CONTENT_204 = 204
HTTP_UNAVAILABLE_503 = 503

# General
RECORDS = "records"
FILE = "_file"
URL = "_url"
BASE64 = "_base64"
IMG_DATA = "_img_data"
COLOR_SPACE = "_color_space"
RESULT_OK = {"status": "OK"}
RESULT_ERROR = {"status": "ERROR", "detail": "Unexpected error."}
ENDPOINT = "https://api.ximilar.com/"

# Services
CUSTOM_IMAGE_RECOGNITION = "custom-recognition"
CUSTOM_OBJECT_DETECTION = "custom-object-detection"
PHOTO_SIMILARITY = "photo-similarity"
PRODUCT_SIMILARITY = "product-similarity"
FASHION_SIMILARITY = "fashion-similarity"
FASHION_SIMPLE = "fashion-search-simple"
HOMEDECOR_SIMILARITY = "homedecor-similarity"
GENERIC_TAGGING = "generic-tagging"
FASHION_TAGGING = "fashion-tagging"
VISUAL_SEARCH = "visual-search"
DOMINANT_COLORS_PRODUCT = "dominant-colors"
DOMINANT_COLORS_GENERIC = "dominant-colors"
HOME_DECOR_TAGGING = "homedecor-tagging"
CUSTOM_SIMILARITY = "custom-similarity"
COLOR_SIMILARITY = "color-similarity"
REMOVEBG = "remove-bg"
IMAGE_MATCHING = "image-matching"
PRODUCTS_MANAGEMENT = "products-management"
COLLECTIBLES_RECOGNITION = "collectibles-recognition"

# Recognition/Vize related
USER_ID = "user_id"
USER = "user"
IMAGE_ID = "image_id"
CREATED = "created"
TYPE_ID = "type_id"
STATUS = "status"
STATUS_OK = "OK"
STATUS_ERROR = "ERROR"
RESULTS = "results"
ID = "id"
NAME = "name"
TASK = "task"
TASKS_COUNT = "tasks_count"
NEGATIVE_FOR_TASK = "negative_for_task"
TYPE = "type"
TASK_TYPE = "type"
LABEL_TYPE = "type"
MULTI_LABEL = "multi_label"
MULTI_CLASS = "multi_class"
REGRESSION = "regression"
DETECTION = "detection"
TAG = "tag"
CATEGORY = "category"
VALUE = "value"
WORKSPACE = "workspace"
IMAGE = "image"
DEFAULT_WORKSPACE = "Default"
IMAGES_COUNT = "images_count"
NORESIZE = "noresize"
LABELS = "labels"
LABEL_NAMES = "label_names"
DETAIL = "detail"
VERSION = "version"
LABEL_ID = "label_id"
TASK_ID = "task_id"
IMG_PATH = "img_path"
THUMB_IMG_PATH = "thumb_img_path"
IMG_WIDTH = "img_width"
IMG_HEIGHT = "img_height"
VERIFY_COUNT = "verifyCount"
TRAIN_STATUS = "train_status"
TASK_NAME = "task_name"
PRODUCTION_VERSION = "production_version"
NEXT = "next"
DATA = "data"
DETECTION_LABEL = "detection_label"
RECOGNITION_LABELS = "recognition_labels"
META_DATA = "meta_data"
DESCRIPTION = "description"
OWNER = "owner"
RECOGNITION_TASKS = "recognition_tasks"
LAST_TRAIN_STATUS = "last_train_status"
OBJECTS = "objects"
BEST_LABEL = "best_label"
TEST_IMAGE = "test_image"
STORE_IMAGES = "store_images"
OUTPUT_NAME = "output_name"
COLOR = "color"
REAL_IMAGE = "real"
VALUES = "values"

# Flows
TOP_NODE = "top_node"
VALID = "valid"

# Similarity search
FIELDS_TO_RETURN = "fields_to_return"
_ID = "_id"
FILTER = "filter"
COUNT = "count"
K_COUNT = "k"
QUERY_RECORD = "query_record"
COLLECTION = "collection"
COLLECTION_ID = "collection-id"
ANSWER_RECORDS = "answer_records"
